<?php
class TagCloudModel extends Model {
	
	private $centroids;
	private $cloud;
	private $clusters;
	
	function TagCloudModel() {
		parent::Model();
	}
	
	/**
	 * Возвращает данные, необходимые для построения облака тегов
	 * (без разбиения на кластеры)
	 *
	 * @return массив с облаком
	 */
	function getTagCloudData() {
		$qGetCloud = "SELECT tags.tag, COUNT(posts_tags.tagid) AS posts_count".
					" FROM posts_tags LEFT JOIN tags ON posts_tags.tagid=tags.id".
					" GROUP BY tags.id";
		$res = $this->db->query($qGetCloud);
		if ($res->num_rows() == 0) {
			return false;
		}
		else {
			return $res->result_array();
		}
	}
	
	/**
	 * Разбивает облако тегов на заданное количество кластеров.
	 *
	 * @param $clustersNum количество кластеров
	 * @return массив с результатами
	 * 	$res[0] - первый кластер, содержит массив со всеми рубриками, которые в него входят
	 * 	вывод записей для всех кластеров
	 * 	foreach ($res as $cluster) {
	 * 		foreach ($cluster as $tag) {
	 * 			echo $tag['tag'].'('.$tag['posts_count'].')';
	 * 		}
	 * 	}
	 */
	function getClusteredCloud($clustersNum) {
		//получаем общее облако тегов
		$this->cloud = $this->getTagCloudData();
		//разбиваем его на заданное количество кластеров
		//определяем начальные центры масс каждого кластера
		//для этого находим теги с максимальным и минимальным количеством постов
		$min = $this->cloud[0]['posts_count'];
		$max = $this->cloud[0]['posts_count'];
		for ($i = 1; $i < count($this->cloud); $i++) {
			if ($this->cloud[$i]['posts_count'] > $max) {
				$max = $this->cloud[$i]['posts_count'];
			}
			if ($this->cloud[$i]['posts_count'] < $min) {
				$min = $this->cloud[$i]['posts_count'];
			}
		}
		//формируем массив с начальными центрами масс
		//идея такая: первый центр масс равен минимальному элементу,
		//последний - максимальному, все остальные - равномерно распределены
		//между ними
		$step = ($max - $min) / ($clustersNum - 1);
		$this->centroids = array();
		for ($i = 0; $i < $clustersNum; $i++) {
			$this->centroids[$i] = $min + $i*$step;
		}
		//первоначальная разбивка на группы
		$prevClusters = array();
		while (TRUE) {
			//рассчитываем расстояние от каждого тега до центра масс
			$distances = $this->_getDistances();
			$this->clusters = $this->_clasterizeTags($distances);
			if ($prevClusters === $this->clusters) {
				break;
			}
			else {
				$prevClusters = $this->clusters;
				$this->centroids = $this->_recalcCentroids();
			}
		}
		//формируем облако тегов, разбитое на кластеры
		//все кластеры, которые не содержат тегов, будут пропущены
		foreach ($this->clusters as $cIndex => $cluster) {
			$cur = 0;
			foreach ($cluster as $tIndex => $tag) {
				if ($tag === 1) {
					$clusteredCloud[$cIndex][$cur] = $this->cloud[$tIndex];
					$cur++;
				}
			}
		}
		return $clusteredCloud;
	}
	
	/**
	 * Определяет расстояния от центров масс до тегов
	 *
	 * @return массив с расстояниями. Его размер [ N x M ], где
	 * 		N - количество элементов в массиве $this->$centroids
	 * 		M - количество элементов в массиве $this->cloud
	 */
	function _getDistances() {
		$distancies = array();
		foreach ($this->centroids as $i => $centroid) {
			foreach ($this->cloud as $j => $tag) {
				$distancies[$i][$j] = abs($centroid - $tag['posts_count']);
			}
		}
		return $distancies;
	}
	
	/**
	 * Возвращает массив с разбивкой тегов на кластеры.
	 *
	 * @param $distances массив с расстояниями
	 * @return массив с разбивкой тегов на кластеры. Его размер [ N x M ], где
	 * 		каждая строка соответсвует кластеру,
	 * 		столбец - тегу,
	 * 		если элемент массива [2][1] равен "1", значит первый тег входит во второй кластер
	 */
	function _clasterizeTags($distances) {
		$clasters = array();
		//определяем число строк и столбцов в массиве
		$rowsNum = count($distances);
		$colsNum = count($distances[0]);
		//просматриваем каждый столбец
		//находим в какой колонке находится минимальный элемент
		//и записываем в нее "1"
		for ($j = 0; $j < $colsNum; $j++) {
			$min = $distances[0][$j];
			$minIndex = 0;
			for ($i = 0; $i < $rowsNum; $i++) {
				$clasters[$i][$j] = 0;
				if ($distances[$i][$j] < $min) {
					$minIndex = $i;
				}
			}
			$clasters[$minIndex][$j] = 1;
		}
		return $clasters;
	}
	
	/**
	 * Пересчитывает центры масс на основе разбивки на группы
	 *
	 * @return новый массив $this->centroids
	 */
	function _recalcCentroids() {
		$newCentroids = array();
		for ($centrIndex = 0; $centrIndex < count($this->centroids); $centrIndex++) {
			$newCentroids[$centrIndex] = 0;
			$n = 0;
			foreach ($this->clusters[$centrIndex] as $key => $val) {
				if ($val === 1) {
					$newCentroids[$centrIndex] += $this->cloud[$key]['posts_count'];
					$n++;
				}
			}
			if ($n > 0) {
				$newCentroids[$centrIndex] /= $n;
			}
			else {
				//оставляем старый центр масс, если в группу не вошел ни один тег
				$newCentroids[$centrIndex] = $this->centroids[$centrIndex];
			}
		}
		return $newCentroids;
	}
}
?>